import datetime
from traceback import format_tb
from typing import List, Any

from webexception.webexception import WebException

from eventix.pydantic.task import TaskModel
from pydantic_db_backend.backend import Backend
from pydantic_db_backend.exceptions import RevisionConflict

# post
# already exists
# still scheduled -> update

# not scheduled --> post new.
# scheduled -> update
#
#   update:
#       get task
#       scheck scheduled, if not post
#       update task
#
#       if worker grabs , version should conflict
#       if conflict: try again


import logging

from pydantic_db_backend.utils import utcnow

log = logging.getLogger(__name__)


def task_post(task: TaskModel) -> TaskModel:
    is_unique = task.unique_key is not None

    if not is_unique:
        # not unique , just try to save. If exists, raise error
        # noinspection PyTypeChecker
        return Backend.post_instance(task)

    # has unique_key

    while True:

        # noinspection PyTypeChecker
        existing_tasks: List[TaskModel] = Backend.get_instances(TaskModel, 0, 10, {"unique_key": task.unique_key})
        next_scheduled_task = next(filter(lambda t: t.status == "scheduled", existing_tasks), None)

        if next_scheduled_task is None:
            # no existing ones that are only scheduled, we have to post
            # noinspection PyTypeChecker
            return Backend.post_instance(task)

        #   update:
        #       get task
        #       scheck scheduled, if not post
        #       update task
        #
        #       if worker grabs , version should conflict
        #       if conflict: try again

        next_scheduled_task.unique_update_from(task)
        try:
            updated_task = Backend.put_instance(next_scheduled_task)
            # noinspection PyTypeChecker
            log.debug(f"updated task {updated_task.uid}")
            # noinspection PyTypeChecker
            return updated_task  # update worked
        except RevisionConflict as e:
            continue  # try again.


def task_clean_expired_workers(namespace: str):
    params = dict(
        model=TaskModel,
        skip=0,
        limit=1,
        query_filter=dict(
            namespace=namespace,  # namespace has to match
            worker_expires={
                "$and": [
                    {"$ne": None},
                    {"$lt": utcnow()},
                ]
            }  # no worker assigned
        ),
        sort=[
            {"priority": "asc"},
            {"eta": "asc"}
        ]
    )

    while True:
        # noinspection PyTypeChecker
        existing_task: TaskModel | None = next(iter(Backend.get_instances(**params)), None)

        # repeat until we were able to take something or nothing is left.
        if existing_task is None:
            break

        existing_task.status = "scheduled"
        existing_task.worker_id = None
        existing_task.worker_expires = None

        try:
            Backend.put_instance(existing_task)
            log.info(f"Released task {existing_task.uid}")
            # noinspection PyTypeChecker
        except RevisionConflict as e:
            continue


def task_next_scheduled(worker_id: str, namespace: str, expires: int = 300) -> TaskModel | None:
    log.debug(f"[{worker_id}] Worker getting next scheduled task...")

    # clean worker expires

    task_clean_expired_workers(namespace)

    # looking up possible tasks in right order
    # take first one
    # try to set worker_id and expiration

    eta = utcnow().isoformat()  # eta has to be now or in the past

    query_filter = dict(
        namespace=namespace,  # namespace has to match
        worker_id=None,  # no worker assigned
        status={"$in": ["scheduled", "retry"]},
        eta={"$lte": eta}
    )
    sort = [
        {"priority": "asc"},
        {"eta": "asc"}
    ]

    while True:  # repeat until we were able to take something or nothing is left.

        # noinspection PyTypeChecker
        existing_task: TaskModel | None = next(iter(Backend.get_instances(
            TaskModel,
            0,
            1,
            query_filter=query_filter,
            sort=sort
        )), None)

        if existing_task is None:
            return None  # no task left

        existing_task.status = "processing"
        existing_task.worker_id = worker_id
        existing_task.worker_expires = utcnow() + datetime.timedelta(seconds=expires)

        try:
            # noinspection PyTypeChecker
            t: TaskModel = Backend.put_instance(existing_task)
            return t
        except RevisionConflict as e:
            continue


def task_set_error(task: TaskModel, error: Any):
    if isinstance(error, WebException):
        task.result = error.dict()
    elif isinstance(error, Exception):
        task.result = dict(
            error_class=error.__class__.__name__,
            error_message=str(error),
            error_status_code=500,
            error_traceback=format_tb(error.__traceback__),
            error_payload={}
        )
    else:
        task.result = error

    if task.retry and (task.max_retries is None or task.max_retries != 0):
        task.status = "retry"
        if task.max_retries is not None:
            task.max_retries -= 1  # decrease max_retries until it reaches zero.

        task.eta = utcnow() + datetime.timedelta(seconds=task.error_eta_inc)
        task.error_eta_inc = min([task.error_eta_inc * 2, task.error_eta_max])

        task.worker_id = None

    else:
        task.status = "error"
        if task.error_expires is not None:
            task.expires = utcnow() + datetime.timedelta(seconds=task.error_expires)

    task.worker_expires = None  # remove worker_expires for prevent cleanup


def task_set_result(task: TaskModel, result: Any):
    task.status = "done"
    if task.store_result:
        task.result = result
        if task.result_expires is not None:
            task.expires = utcnow() + datetime.timedelta(seconds=task.result_expires)
    else:
        task.expires = utcnow()

    task.worker_expires = None  # remove worker_expires for prevent cleanup




import logging

from fastapi import APIRouter

from eventix.pydantic.event import EventModel

log = logging.getLogger(__name__)

router = APIRouter(tags=["event"])


@router.get("/events")
async def events_get():
    return {}


@router.post("/event")
async def events_post(event: EventModel) -> EventModel:
    return event


@router.get("/event/{uid}")
async def events_get(uid: str) -> EventModel:
    return EventModel()


@router.delete("/event/{uid}")
async def events_delete(uid: str) -> None:
    return None



import datetime

from pydantic.fields import Field

from pydantic_db_backend.backend import BackendModel
from pydantic_db_backend.utils import utcnow


class EventModel(BackendModel):
    payload: dict | None = Field(default_factory=dict)
    schedule_time: datetime.datetime | None = Field(default_factory=utcnow)
    unique_key: str | None = None
    priority: int | None = 0





import traceback
from typing import Dict, Optional, Any, List, Type
from fastapi import Request, HTTPException
from starlette.middleware.base import BaseHTTPMiddleware, DispatchFunction, RequestResponseEndpoint
from starlette.responses import JSONResponse, Response

import logging

from starlette.types import ASGIApp
from webexception.webexception import WebException

from pydantic_db_backend.backend import Backend, backend_context_var, backend_alias_context_var, BackendBase
from pydantic_db_backend.backends.couchdb import CouchDbBackend
from pydantic_db_backend.exceptions import BackendException

log = logging.getLogger(__name__)


class BackendMiddleware(BaseHTTPMiddleware):
    def __init__(
        self,
        app: ASGIApp,
        dispatch: DispatchFunction | None = None,
        backend: Type[BackendBase] = None
    ) -> None:
        super().__init__(app, dispatch)
        self.backend = backend

    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
        with Backend.provider(self.backend):
            response = await call_next(request)
        return response


class ErrorMiddleware(BaseHTTPMiddleware):
    def __init__(
        self,
        app: ASGIApp,
        dispatch: DispatchFunction | None = None,
    ) -> None:
        super().__init__(app, dispatch)

    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
        try:
            response = await call_next(request)
        except Exception as e:
            return await self.make_exception_response(e)
        return response

    async def make_exception_detail(self, e: Exception) -> Dict:
        tb = traceback.format_exc()
        ret = dict(error_class=e.__class__.__name__, error_message=str(e), error_traceback=tb)
        return ret

    async def make_exception_response(self, e: Exception, headers: Optional[Dict[str, Any]] = None) -> JSONResponse:
        status_code = None
        if isinstance(e, HTTPException):
            status_code = e.status_code
            content = dict(detail=e.detail)

        elif isinstance(e, WebException):
            status_code = e.status_code
            if status_code == 204:
                content = None
            else:
                content = e.dict()
        else:
            if status_code is None:  # no status code found yet
                status_code = 500
            content = dict(detail=await self.make_exception_detail(e))

        if status_code >= 500:
            log.exception(e, stacklevel=2)

        if isinstance(content, dict):
            response = JSONResponse(status_code=status_code, content=content, headers=headers)
        else:
            response = Response(status_code=status_code, headers=headers, content=content)
        return response



from __future__ import annotations

import datetime
from typing import Any, Dict, Tuple, Self, Type, Literal

from pydantic import Field
from webexception.webexception import WebException

from pydantic_db_backend.backend import BackendModel
from pydantic_db_backend.utils import utcnow, _uid

task_model_status = Literal["scheduled", "processing", "done", "error", "retry"]


class TaskModel(BackendModel):
    unique_key: str | None = None
    eta: datetime.datetime | None = Field(default_factory=utcnow)
    scheduled: datetime.datetime | None = Field(default_factory=utcnow)
    status: task_model_status | None = "scheduled"
    priority: int | None = 0  # lower value -> higher priority, because of sorting
    namespace: str | None = "default"
    task: str
    args: tuple | None = Field(default_factory=tuple)
    kwargs: Dict[str, Any] | None = Field(default_factory=dict)

    worker_id: str | None = None
    worker_expires: datetime.datetime | None = None

    retry: bool = True
    max_retries: int | None = None  # None forever, positive value gets decremented until 0
    error_eta_inc: int = 15
    error_eta_max: int = 300
    store_result: bool = True

    result: Any = None

    # Expirations will keep the task in the db.
    # If there is one with None it will keep forever.
    # Otherwise the expiration is set to the one with the longest duration
    # depending which ones are applicable.

    error_expires: int | None = None  # keep forever on error
    result_expires: int | None = 604800  # 7 days

    expires: datetime.datetime | None = None

    def unique_update_from(self, task: Self):
        self.eta = min([task.eta, self.eta])
        self.priority = min([task.priority, self.priority])

    class Config:
        backend_indexes = [
            ("priority_eta", [{"priority": "asc"}, {"eta": "asc"}]),
        ]

    # sort order
    # priority-, eta-

    # 9    0
    # 9    1
    # 9    2
    # 9    3
    # 8    1
    # 8    2
    # 5    4



import contextlib
from typing import List, Type

import pydash
from fastapi import HTTPException


@contextlib.contextmanager
def raise_errors(r, exceptions: List[Type[Exception]]):
    exceptions_by_class = {e.__name__: e for e in exceptions}
    if r.status_code == 200:
        yield r
    else:
        json = r.json()
        detail = pydash.get(json, 'detail', json)
        error_class = pydash.get(detail, 'error_class')
        payload = pydash.get(detail, 'error_payload', {})
        if error_class in exceptions_by_class:
            e = exceptions_by_class[error_class](**payload)
            raise e
        # backend errors
        raise HTTPException(status_code=r.status_code, detail=detail)



from __future__ import annotations

from typing import Any

from requests import Session, Response

import logging

log = logging.getLogger(__name__)


class EventixClientSession(Session):
	def __init__(self, base_url: str = None) -> None:
		self.base_url = base_url
		super().__init__()

	def request(
		self,
		method: str | bytes,
		url: str | bytes,
		*args,
		**kwargs
	) -> Response:  # pragma: no cover
		return super().request(
			method,
			f"{self.base_url}{url}",
			*args,
			**kwargs
		)


class EventixClient:
	interface: Any | None = EventixClientSession()
	namespace: str | None = None

	@classmethod
	def set_base_url(cls, base_url):
		if isinstance(cls.interface, EventixClientSession):
			log.info(f"Setting EventixClient base_url: {base_url}")
			cls.interface.base_url = base_url



import logging

from eventix.exceptions import backend_exceptions
from eventix.functions.errors import raise_errors
from eventix.functions.eventix_client import EventixClient
from eventix.pydantic.task import TaskModel

log = logging.getLogger(__name__)


class TaskScheduler(EventixClient):

    @classmethod
    def schedule(cls, task: TaskModel) -> TaskModel:
        log.debug(f"scheduling task: {task.task} uid: {task.uid} eta: {task.eta} unique_key: {task.unique_key}")
        task.namespace = cls.namespace
        return cls.task_post(task)

    @classmethod
    def task_get(cls, uid: str) -> TaskModel:
        r = cls.interface.get(f'/task/{uid}')
        with raise_errors(r, backend_exceptions):
            return TaskModel.parse_raw(r.content)

    @classmethod
    def task_post(cls, task: TaskModel) -> TaskModel:
        r = cls.interface.post('/task', data=task.json())
        with raise_errors(r, backend_exceptions):
            return TaskModel.parse_raw(r.content)



from __future__ import annotations

import importlib
import logging
import os
import sys
import time
from typing import List

import dotenv
from pydantic_db_backend.utils import utcnow

from eventix.exceptions import TaskNotRegistered, backend_exceptions
from eventix.functions.core import EventflowTaskBase, namespace_context, worker_id_context, namespace_provider
from eventix.functions.errors import raise_errors
from eventix.functions.eventix_client import EventixClient
from eventix.functions.task import task_set_error, task_set_result
from eventix.pydantic.task import TaskModel

log = logging.getLogger(__name__)


class TaskWorker(EventixClient):
    _wait_interval = 10
    _tasks = {}
    namespace: str = None

    @classmethod
    def setup_logging(cls, without_time: bool = False, level: int = logging.INFO):
        if without_time:
            f = "[%(levelname)8s] %(message)s"
        else:
            f = "%(asctime)s [%(levelname)s] %(message)s"
        logging.basicConfig(
            level=level,
            format=f,
            handlers=[
                logging.StreamHandler()
            ]
        )

    @classmethod
    def config(cls, config: dict):

        base_url = os.environ.get("EVENTIX_URL", "")
        if base_url == "":
            log.error("No EVENTIX_URL set.")
            sys.exit()

        cls.set_base_url(base_url)

        if "register_tasks" in config:
            cls.register_tasks(config['register_tasks'])

        if "namespace" in config:
            cls.namespace = config['namespace']

        namespace = os.environ.get("EVENTIX_NAMESPACE", "")
        if namespace != "":
            cls.namespace = namespace

    @classmethod
    def register_tasks(cls, paths=List[str]):
        log.info("registering tasks...")
        # noinspection PyTypeChecker
        for path in paths:
            try:
                imported_module = importlib.import_module(path)
                for f in filter(
                    lambda x: isinstance(x, EventflowTaskBase),
                    [getattr(imported_module, x) for x in dir(imported_module)]
                ):
                    log.info(f"registered '{f.func_name}' from {path}")
                    cls._tasks[f.func_name] = f
            except ImportError as e:
                print(e)

    @classmethod
    def task_next_scheduled(cls) -> TaskModel | None:
        with namespace_context() as namespace:
            with worker_id_context() as worker_id:
                params = dict(
                    worker_id=worker_id,
                    namespace=namespace
                )
                r = cls.interface.get(f'/task/next_scheduled', params=params)
                if r.status_code == 200:
                    tm = TaskModel.parse_raw(r.content)
                    return tm
                else:
                    return None

    # with raise_errors(r):
    #     return TaskModel.parse_raw(r.content)

    @classmethod
    def listen(cls, endless=True):
        log.info("Start listening...")
        while True:
            log.info("Looking for tasks...")
            t = cls.task_next_scheduled()
            if t is not None:
                cls.execute_task(t)
            else:
                log.info(f"Nothing to do... waiting {cls._wait_interval}s")
                if not endless:
                    return
                time.sleep(cls._wait_interval)

    @classmethod
    def execute_task(cls, task: TaskModel):
        try:
            log.info(f"Executing task: {task.task} uid: {task.uid} ...")

            if task.task not in cls._tasks:
                raise TaskNotRegistered(task=task.task)  # Task not registered in worker

            f = cls._tasks[task.task]
            r = f.run(*task.args, **task.kwargs)
            if task.store_result:
                task_set_result(task, r)

        except TaskNotRegistered as e:
            log.exception(e)
            task_set_error(
                task,
                TaskNotRegistered(task=task.task)
            )

        except Exception as e:
            log.exception(e)
            task_set_error(task, e)

        finally:
            cls.task_write_back(task)

    @classmethod
    def task_write_back(cls, task: TaskModel) -> TaskModel | None:
        try:

            if task.expires is not None and task.expires < utcnow():  # gammel
                # if task is already expired, delete it instead of updating
                r = cls.interface.delete(f'/task/{task.uid}')
                return None
            else:
                # task not yet expired.... update
                r = cls.interface.put(f'/task/{task.uid}', data=task.json())

            with raise_errors(r, backend_exceptions):
                if r.status_code == 200:
                    tm = TaskModel.parse_raw(r.content)
                    return tm

        except Exception as e:
            log.error("Exception raised when calling eventix")
            log.exception(e)
            log.error("Exception used this task info")
            log.error(task.json())

        return None

    @classmethod
    def load_env(cls):
        dotenv.load_dotenv(".env.local")

    def __init__(self, config: dict) -> None:
        self.setup_logging()
        self.load_env()
        self.config(config)

    def start(self, endless: bool = True):
        log.info(f"Using namespace: {self.namespace}")
        with namespace_provider(self.namespace):
            self.listen(endless)



import datetime
import json
from uuid import uuid4

import psutil


def uid() -> str:
    return str(uuid4()).replace("-", "")


_uid = uid


def metrics() -> dict:
    memory_info = psutil.Process().memory_info()
    return {
        "memory": dict(rss=memory_info.rss, vms=memory_info.vms, data=memory_info.data)
    }




from pydantic_db_backend.exceptions import AlreadyExists, NotFound, RevisionConflict
from webexception.webexception import WebException


class TaskNotUnique(WebException):
    status_code = 409

    def __init__(self, uid: str) -> None:
        super().__init__(
            f"Task with uid '{uid}' already exists and task unique is not activated for overwriting.",
            uid=uid
        )


class NoTaskFound(WebException):
    status_code = 204

    def __init__(self, namespace: str) -> None:
        super().__init__(
            f"No task for namespace {namespace}",
            namespace=namespace
        )


class TaskNotRegistered(WebException):
    status_code = 404

    def __init__(self, task: str) -> None:
        super().__init__(
            f"Task '{task}' not registered.",
            task=task
        )



backend_exceptions = [
    AlreadyExists,
    NotFound,
    RevisionConflict
]




import contextlib
import contextvars
import functools
import logging
import os
from inspect import signature, Parameter
from typing import Callable, Tuple, Dict, Any, List

from pydantic import BaseModel

from eventix.functions.task_scheduler import TaskScheduler
from eventix.pydantic.task import TaskModel

log = logging.getLogger(__name__)


class EventflowTaskBase(object):
    pass


namespace_context_var = contextvars.ContextVar('namespace_context_var', default=None)


@contextlib.contextmanager
def namespace_provider(namespace: str):
    # noinspection PyTypeChecker
    token = namespace_context_var.set(namespace)
    yield
    namespace_context_var.reset(token)


@contextlib.contextmanager
def namespace_context():
    # noinspection PyTypeChecker
    namespace = namespace_context_var.get()
    if namespace is None:
        namespace = os.environ.get("EVENTFLOW_TASK_NAMESPACE", "default")
    yield namespace


worker_id_context_var = contextvars.ContextVar('worker_id_context_var', default=None)


@contextlib.contextmanager
def worker_id_provider(worker_id: str):
    # noinspection PyTypeChecker
    token = worker_id_context_var.set(worker_id)
    yield
    worker_id_context_var.reset(token)


@contextlib.contextmanager
def worker_id_context():
    # noinspection PyTypeChecker
    worker_id = worker_id_context_var.get()
    if worker_id is None:
        worker_id = os.environ.get("EVENTFLOW_TASK_WORKER_ID", "default")
    yield worker_id


def task(store_result: bool = True, unique_key_generator: Callable = None):
    # parameters suggested:
    #
    # store_result: bool , wether to store or not to store results
    # unique_uid: ....  having a unique id, which leads to update the task if it is rescheduled.

    # retry: bool, wether to retry failed tasks
    # max_retry: int, maximum retries default 5 ... each retry doubles the time to wait, starting with 30 sec

    is_unique = unique_key_generator is not None

    def inner(f):
        class EventflowTask(EventflowTaskBase):
            def __init__(self):
                self.func = f
                self.func_name = f.__name__
                self.arg_spec = list(signature(f).parameters.values())
                self.arg_spec_args = [arg.annotation for arg in self.arg_spec]
                self.arg_spec_kwargs = {arg.name: arg.annotation for arg in self.arg_spec}
                assert True

            def restore_pydantic_instances(self, args, kwargs) -> tuple[list[Any], Dict[str, Any]]:
                new_args = []
                for i, arg in enumerate(args):
                    if isinstance(arg, dict):
                        cls = self.arg_spec_args[i]
                        if issubclass(cls, BaseModel):
                            arg = cls.parse_obj(arg)
                    new_args.append(arg)
                new_kwargs = {}
                for kw, arg in kwargs.items():
                    if isinstance(arg, dict):
                        cls = self.arg_spec_kwargs[kw]
                        if issubclass(cls, BaseModel):
                            arg = cls.parse_obj(arg)
                    new_kwargs[kw] = arg
                return new_args, new_kwargs

            @functools.wraps(f)
            def delay(self, *args, _priority: int | None = 0, **kwargs):
                # priority has to be negated, needed for fixing ascending sort order

                tm = self.make_task_model(args, kwargs, priority=_priority * -1)
                # log.debug(f"scheduling {self.func.__name__} info: {tm.json()}")
                tm = TaskScheduler.schedule(tm)
                return tm

            @functools.wraps(f)
            def run(self, *args, **kwargs):
                args, kwargs = self.restore_pydantic_instances(args, kwargs)
                return self.func(*args, **kwargs)

            def make_task_model(self, args, kwargs, priority) -> TaskModel:
                params = dict(
                    task=self.func_name,
                    args=args,
                    kwargs=kwargs,
                    priority=priority,
                    store_result=store_result
                )
                if is_unique:
                    unique_key = unique_key_generator(*args, **kwargs)
                    params |= dict(unique_key=unique_key)

                return TaskModel.parse_obj(params)

        return EventflowTask()

    return inner
